/* talk.c    conversation with checkers engines  */

/*
 * Dame:     a point and click checkerboard
 * (C):      1998  Peter Chiocchetti  <girbal@super.tacheles.de>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define TRIGGER "chooses "


#include <dame.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/wait.h>


static int trig;

static FILE *to_black;
static FILE *from_black;
static FILE *to_white;
static FILE *from_white;

static pid_t black_pid = -1;
static pid_t white_pid = -1;

static guint black_tag, white_tag;

/*
 * How are you ?
 */

/* Forward move to an engine !  */
void
tell_move (char move[MAXDEPTH * 5 + 1])
{
  FILE *to_fd = (players.current == Black) ? to_black : to_white;

  if (!game_over) {
    fprintf (to_fd, "%s\n", move);
    fflush  (to_fd);
  }
}

/* Split engine's output into calls to moving function.  */
static void
listen_move (gpointer data, gint source, GdkInputCondition condition)
{
  FILE   *from_fd = data;
  char   *buf = NULL;
  size_t  len = 0;
  char   *pat;
  char   *tail;
  int     home;
  int     target;

  /* getline sets EOF error, when engine prompts for input.  */
  if (ferror (from_fd))
    clearerr (from_fd);

  /* Read in a line from the engine.  */
  while (!game_over && getline (&buf, &len, from_fd) > 0) {
/*     fprintf (stderr, "%s", buf); */
    /* Does it contain the pattern ?  */
    if ((pat = strstr (buf, TRIGGER))) {
      /* Parse it.  */
      pat += trig;
      target = strtol (pat, &tail, 0);
      pat = tail;

      while (!isspace (*pat)) {
	/* There is more nonwhitespace,  */
	if (ispunct (*pat)) {
	  home = target;
	  pat++;
	  continue;
	}
	/* so it ought to be another number.  */
	target = strtol (pat, &tail, 0);
	trace_move (home, target);
	/* Advance past it.  */
	pat = tail;
      }
    }
  }
  free (buf);
}

void
stop_talk (void)
{
  int status;

  /* Clear Signal handler  */
  signal (SIGPIPE, SIG_IGN);
  signal (SIGCHLD, SIG_IGN);

  if (black_tag) {
    gdk_input_remove (black_tag);
    black_tag = 0;
  }
  if (black_pid != -1) {
    fclose (from_black);
    fclose (to_black);
    if (!kill (black_pid, SIGKILL))
      waitpid (black_pid, &status, 0);
    black_pid = -1;
  }

  if (white_tag) {
    gdk_input_remove (white_tag);
    white_tag = 0;
  }
  if (white_pid != -1) {
    fclose (from_white);
    fclose (to_white);
    if (!kill (white_pid, SIGKILL))
      waitpid (white_pid, &status, 0);
    white_pid = -1;
  }
}


/* 
 * Handle [SIGCHLD and SIGPIPE]
 */

void
catch_signal (int signum)
{
  game_over = 1;
  stop_talk ();
  fprintf (stderr, ("\nDame : "));
  fprintf (stderr, ("%s\n", g_strsignal (signum)));
  gnome_appbar_set_status (GNOME_APPBAR (ab), g_strsignal (signum));
/*   psignal (signum, NULL); */
}


/* Computer players: The player's name is its command line invocation.
   No parameters: First input is the _opponents_ colour (engine asks,
   which colour do YOU want to play?), second the strength level. This
   is some kind of protocol: b means black, w means white; levels are
   from one to nine (1 - 9). Engine must take these, and may further
   ignore them.  */

void
start_talk ()
{
  int from_pipe [2];
  int to_pipe[2];
  int status_flags;
  black_tag = 0;
  white_tag = 0;

  /* Dead Engines cause havoc !  */
  signal (SIGPIPE, catch_signal);
  signal (SIGCHLD, catch_signal);

  /* Start black player program.  */
  if (strcmp (players.black, username)) {
    /* Create the pipe. */
    if ((pipe (from_pipe) < 0) || (pipe(to_pipe) < 0)) {
      fprintf (stderr, "Dame : Pipe failed.\n");
      exit (1);
    }

    /* Create the child process. */
    if ((black_pid = fork ()) == 0) {
      /* This is the child process.  */
      status_flags = fcntl (from_pipe[0], F_GETFL, 0);
      status_flags |= O_NONBLOCK;
      fcntl(from_pipe[0], F_SETFL, status_flags);

      dup2 (to_pipe  [0], STDIN_FILENO);
      dup2 (from_pipe[1], STDOUT_FILENO);

      close (to_pipe  [0]);
      close (to_pipe  [1]);
      close (from_pipe[0]);
      close (from_pipe[1]);

      execlp (players.black, players.black, NULL);

      fprintf (stderr, "Dame : can't exec %s.\n", players.black);
      exit (1);
    }
    if (black_pid < 0) {
      /* The fork failed.  */
      fprintf (stderr, "Dame : fork failed.\n");
      stop_talk ();
      exit (1);
      }
    
    /* Install pipe with black player.  */
    else {
      /* This is the parent process.  */
      close (to_pipe  [0]);
      close (from_pipe[1]);

      from_black = fdopen (from_pipe[0], "r");
      to_black   = fdopen (to_pipe  [1], "w");

      setvbuf (from_black, NULL, _IONBF, 0);
      setvbuf (to_black,   NULL, _IONBF, 0);

      black_tag = gdk_input_add (fileno (from_black), GDK_INPUT_READ,
				 (GdkInputFunction) listen_move, from_black);

      fprintf (to_black, "w\n");
      fflush (to_black);
      fprintf (to_black, "%s\n", players.black_level);
      fflush (to_black);
    }
  }

  /* Start white player program.  */
  if (strcmp (players.white, username)) {
    /* Create the pipe. */
    if ((pipe (from_pipe) < 0) || (pipe (to_pipe) < 0)) {
      fprintf (stderr, "Dame : Pipe failed.\n");
      exit (1);
    }

    /* Create the child process. */
    if ((white_pid = fork ()) == 0) {
      /* This is the child process.  */
      status_flags = fcntl (from_pipe[0], F_GETFL, 0);
      status_flags |= O_NONBLOCK;
      fcntl(from_pipe[0], F_SETFL, status_flags);

      dup2 (to_pipe  [0], STDIN_FILENO);
      dup2 (from_pipe[1], STDOUT_FILENO);

      close (to_pipe  [0]);
      close (to_pipe  [1]);
      close (from_pipe[0]);
      close (from_pipe[1]);

      execlp (players.white, players.white, NULL);

      fprintf (stderr, "Dame : can't exec %s.\n", players.white);
      exit (1);
    }
    if (white_pid < 0) {
      /* The fork failed.  */
      fprintf (stderr, "Dame : fork failed.\n");
      stop_talk ();
      exit (1);
      }

    /* Install pipe with white player.  */
    else {
      /* This is the parent process.  */
      close (to_pipe  [0]);
      close (from_pipe[1]);

      from_white = fdopen (from_pipe[0], "r");
      to_white   = fdopen (to_pipe  [1], "w");

      setvbuf (from_white, NULL, _IONBF, 0);
      setvbuf (to_white,   NULL, _IONBF, 0);

      white_tag = gdk_input_add (fileno (from_white), GDK_INPUT_READ,
				 (GdkInputFunction) listen_move, from_white);

      fprintf (to_white, "b\n");
      fflush (to_white);
      fprintf (to_white, "%s\n", players.white_level);
      fflush (to_white);
    }
  }
  /* Prepare for listen_move.  */
  trig = strlen (TRIGGER);
}
